from __future__ import absolute_import
from __future__ import unicode_literals
import collections
from functools import total_ordering

from sqlalchemy import bindparam, text, column
from sqlalchemy.sql import operators, and_, or_, not_

from sqlagg.exceptions import SqlAggException


class NotEqMixin(object):
    def __ne__(self, other):
        return not self.__eq__(other)


@total_ordering
class SqlFilter(NotEqMixin):
    def build_expression(self):
        raise NotImplementedError()

    def __lt__(self, other):
        """Ordering is required for consistent sorting when creating column keys"""
        return hash(self) < hash(other)


class RawFilter(SqlFilter):
    def __init__(self, expression):
        self.expression = expression

    def build_expression(self):
        return text(self.expression)

    def __eq__(self, other):
        return isinstance(other, RawFilter) and self.expression == other.expression

    def __hash__(self):
        return hash((RawFilter, self.expression))

    def __repr__(self):
        return "SQL({})".format(self.expression)


class BasicFilter(SqlFilter):
    operator = None
    operator_string = None

    def __init__(self, column_name, parameter, operator=None):
        self.column_name = column_name
        self.parameter = parameter
        if operator:
            self.operator = operator

    def build_expression(self):
        if not self.operator:
            raise SqlAggException('Operator missing')

        return self.operator(column(self.column_name), bindparam(self.parameter))

    def __eq__(self, other):
        return (
            isinstance(other, BasicFilter)
            and self.column_name == other.column_name
            and self.parameter == other.parameter
            and self.operator == other.operator
        )

    def __hash__(self):
        return hash((type(self), self.column_name, self.parameter, self.operator))

    def __repr__(self):
        return "SQL({} {} {})".format(self.column_name, self.operator_string, self.parameter)


class BetweenFilter(SqlFilter):
    def __init__(self, column_name, lower_param, upper_param):
        self.column_name = column_name
        self.lower_param = lower_param
        self.upper_param = upper_param

    def build_expression(self):
        return column(self.column_name).between(
            bindparam(self.lower_param), bindparam(self.upper_param)
        )

    def __eq__(self, other):
        return (
            isinstance(other, BetweenFilter)
            and self.column_name == other.column_name
            and self.lower_param == other.lower_param
            and self.upper_param == other.upper_param
        )
    def __hash__(self):
        return hash((BetweenFilter, self.column_name, self.lower_param, self.upper_param))


class GTFilter(BasicFilter):
    operator = operators.gt
    operator_string = '>'


class GTEFilter(BasicFilter):
    operator = operators.ge
    operator_string = '>='


class LTFilter(BasicFilter):
    operator = operators.lt
    operator_string = '<'


class LTEFilter(BasicFilter):
    operator = operators.le
    operator_string = '<='


class EQFilter(BasicFilter):
    operator = operators.eq
    operator_string = '='


class NOTEQFilter(BasicFilter):
    operator = operators.ne
    operator_string = '!='


class INFilter(BasicFilter):
    """
    This filter requires that the parameter value be a tuple.
    """
    operator_string = 'in'
    def build_expression(self):
        assert isinstance(self.parameter, collections.Iterable)
        return operators.in_op(
            column(self.column_name),
            tuple(bindparam(param) for param in self.parameter)
        )

    def __eq__(self, other):
        return (
            isinstance(other, INFilter)
            and self.column_name == other.column_name
            and self.operator == other.operator
            and sorted(self.parameter) == sorted(other.parameter)
        )

    def __hash__(self):
        return hash((type(self), self.column_name, self.operator, tuple(sorted(self.parameter))))


class ISNULLFilter(SqlFilter):
    def __init__(self, column_name):
        self.column_name = column_name

    def build_expression(self):
        return column(self.column_name).is_(None)

    def __eq__(self, other):
        return isinstance(other, ISNULLFilter) and self.column_name == other.column_name

    def __hash__(self):
        return hash((ISNULLFilter, self.column_name))

    def __repr__(self):
        return "SQL({} IS NULL)"


class NOTNULLFilter(SqlFilter):
    def __init__(self, column_name):
        self.column_name = column_name

    def build_expression(self):
        return column(self.column_name).isnot(None)

    def __eq__(self, other):
        return isinstance(other, NOTNULLFilter) and self.column_name == other.column_name

    def __hash__(self):
        return hash((NOTNULLFilter, self.column_name))

    def __repr__(self):
        return "SQL({} NOT NULL)"


class NOTFilter(SqlFilter):
    def __init__(self, filter):
        self.filter = filter

    def build_expression(self):
        return not_(self.filter.build_expression())

    def __eq__(self, other):
        return isinstance(other, NOTFilter) and self.filter == other.filter

    def __hash__(self):
        return hash((NOTFilter, self.filter))


class ANDFilter(SqlFilter):
    """
    Lets you construct AND operations on filters.
    """
    def __init__(self, filters):
        self.filters = filters
        assert len(self.filters) > 1

    def build_expression(self):
        return and_(*[f.build_expression() for f in self.filters])

    def __eq__(self, other):
        return isinstance(other, ANDFilter) and set(self.filters) == set(other.filters)

    def __hash__(self):
        return hash((NOTFilter,) + tuple(sorted(self.filters)))

    def __repr__(self):
        return "SQL({})".format(' AND '.join(self.filters))


class ORFilter(SqlFilter):
    """
    Lets you construct OR operations on filters.
    """
    def __init__(self, filters):
        self.filters = filters
        assert len(self.filters) > 1

    def build_expression(self):
        return or_(*[f.build_expression() for f in self.filters])

    def __eq__(self, other):
        return isinstance(other, ORFilter) and set(self.filters) == set(other.filters)

    def __hash__(self):
        return hash((ORFilter,) + tuple(sorted(self.filters)))

    def __repr__(self):
        return "SQL({})".format(' OR '.join(self.filters))


RAW = RawFilter
BETWEEN = BetweenFilter
GT = GTFilter
GTE = GTEFilter
LT = LTFilter
LTE = LTEFilter
EQ = EQFilter
NOTEQ = NOTEQFilter
IN = INFilter
ISNULL = ISNULLFilter
NOTNULL = NOTNULLFilter
NOT = NOTFilter
AND = ANDFilter
OR = ORFilter



from __future__ import absolute_import
from __future__ import unicode_literals

from sqlalchemy import desc, asc
from sqlalchemy.sql import column


class OrderBy(object):
    def __init__(self, column_name, is_ascending=True):
        self.column_name = column_name
        self.is_ascending = is_ascending

    @property
    def column(self):
        return column(self.column_name)

    def build_expression(self):
        if self.is_ascending:
            return asc(self.column)
        return desc(self.column)

    def __str__(self):
        return (
            'OrderBy(column_name=%s, is_ascending=%s)'
            % (self.column_name, str(self.is_ascending))
        )



from __future__ import absolute_import
from __future__ import unicode_literals
from sqlalchemy import func, distinct, case, text, cast, Integer, column
from sqlalchemy.dialects.postgresql import aggregate_order_by
from .base import BaseColumn, CustomQueryColumn, SqlColumn
import six
import uuid


class SimpleColumn(BaseColumn):
    pass


class YearColumn(BaseColumn):
    aggregate_fn = lambda _, y: func.extract('YEAR', y)


class MonthColumn(BaseColumn):
    aggregate_fn = lambda _, y: func.extract('MONTH', y)


class WeekColumn(BaseColumn):
    aggregate_fn = lambda _, y: func.extract('WEEK', y)


class DayColumn(BaseColumn):
    aggregate_fn = lambda _, y: func.extract('DAY', y)


class YearQuarterColumn(BaseColumn):
    aggregate_fn = lambda _, y: func.extract('QUARTER', y)


class DayOfWeekColumn(BaseColumn):
    aggregate_fn = lambda _, y: func.extract('DOW', y)


class DayOfYearColumn(BaseColumn):
    aggregate_fn = lambda _, y: func.extract('DOY', y)


class SumColumn(BaseColumn):
    aggregate_fn = func.sum


class CountColumn(BaseColumn):
    aggregate_fn = func.count


class MaxColumn(BaseColumn):
    aggregate_fn = func.max


class MinColumn(BaseColumn):
    aggregate_fn = func.min


class MeanColumn(BaseColumn):
    aggregate_fn = func.avg

    def get_value(self, row):
        value = super(MeanColumn, self).get_value(row)
        if value is not None:
            return float(value)


class NonzeroSumColumn(BaseColumn):
    aggregate_fn = lambda _, column: cast(func.sum(column) > 0, Integer)


class CountUniqueColumn(BaseColumn):
    aggregate_fn = lambda _, column: func.count(distinct(column))


class ConditionalAggregation(BaseColumn):
    def __init__(self, key=None, whens=None, else_=None, *args, **kwargs):
        super(ConditionalAggregation, self).__init__(key, *args, **kwargs)
        self.whens = whens or []
        self.else_ = else_

        assert self.key or self.alias, "Column must have either a key or an alias"

    @property
    def sql_column(self):
        return ConditionalColumn(self.key, self.whens, self.else_, self.aggregate_fn, self.alias)


class ArrayAggColumn(BaseColumn):
    """
    Perform array aggregation on a column
    Pass order_by_col to sort by another column within the group
    Example: array generated for a column col1, ordered by col2 with select clause like
    Select ARRAY_AGG(col1 ORDER BY col2), col3 ..
    """

    def __init__(self, key, order_by_col=None, *args, **kwargs):
        super(ArrayAggColumn, self).__init__(key, *args, **kwargs)
        self.order_by_col = order_by_col

    @property
    def sql_column(self):
        return ArrayAggSQLColumn(self.key, self.order_by_col, self.alias)


class SumWhen(ConditionalAggregation):
    """
    Without binds:
    SumWhen("vehicle", whens=[["unicycle", 1], ["bicycle", 2], ["car", 4]], else_=0, alias="num_wheels")

    With binds:
    SumWhen("age_cohort",
            whens=[
                ["age_in_months < ?", 6, 1],
                ["age_in_months < ?", 12, 2],
            ],
            else_=0,
            alias="half_years")
    """
    aggregate_fn = func.sum


class ConditionalColumn(SqlColumn):
    """
    Without binds:
    ConditionalColumn("vehicle",
                      whens=[["unicycle", 1], ["bicycle": 2], ["car": 4]],
                      else_=0,
                      aggregation_fn=func.sum,
                      alias="num_wheels")

    Or with binds:
    ConditionalColumn("age_cohort",
                      whens=[
                        ["age_in_months < ?", 6, 1],
                        ["age_in_months < ?", 12, 2],
                      ],
                      else_=0,
                      aggregation_fn=func.sum)
    """
    def __init__(self, column_name, whens, else_, aggregate_fn, alias):
        self.aggregate_fn = aggregate_fn
        self.column_name = column_name
        self.whens = whens
        self.else_ = else_
        self.alias = alias

    @property
    def label(self):
        return self.alias or self.column_name

    def build_column(self):
        if self.column_name:
            expr = case(value=column(self.column_name), whens=self.whens, else_=self.else_)
        else:
            expr = case(whens=self._build_whens(), else_=self.else_)

        if self.aggregate_fn:
            expr = self.aggregate_fn(expr)
        return expr.label(self.label)

    def _build_whens(self):
        whens = []
        for item in self.whens:
            when, *binds, then = item
            if binds:
                binds = list(reversed(binds))
                named_binds = {}
                when_with_named_binds = ''
                for letter in when:
                    if letter != '?':
                        when_with_named_binds += letter
                    else:
                        bind_name = 'p' + uuid.uuid4().hex
                        when_with_named_binds += ':' + bind_name
                        named_binds[bind_name] = binds.pop()
                when = text(when_with_named_binds).bindparams(**named_binds)
            else:
                when = text(when)
            then = text(then) if isinstance(then, six.string_types) else then
            whens.append((when, then))
        return whens


class ArrayAggSQLColumn(SqlColumn):
    def __init__(self, column_name, order_by_col, alias=None):
        self.column_name = column_name
        self.order_by_col = order_by_col
        self.alias = alias

    @property
    def label(self):
        return self.alias or self.column_name

    def build_column(self):
        table_column = column(self.column_name)
        if self.order_by_col:
            order_by_column = column(self.order_by_col)
            return func.array_agg(aggregate_order_by(table_column, order_by_column.asc())).label(self.label)
        return func.array_agg(table_column).label(self.label)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import collections
from collections import OrderedDict

import sqlalchemy
from sqlalchemy import column, table

from sqlagg.exceptions import ColumnNotFoundException, SqlAggException, \
    DuplicateColumnsException
from sqlagg.filters import SqlFilter
from six.moves import zip


class SqlColumn(object):
    @property
    def label(self):
        raise NotImplementedError()

    def build_column(self):
        raise NotImplementedError()


class SimpleSqlColumn(SqlColumn):
    """
    Simple representation of a column with a name and an aggregation function which can be None.
    """
    def __init__(self, column_name, aggregate_fn=None, alias=None):
        self.column_name = column_name
        self.alias = alias
        self.aggregate_fn = aggregate_fn

    @property
    def label(self):
        return self.alias or self.column_name

    def build_column(self):
        table_column = column(self.column_name)
        sql_col = self.aggregate_fn(table_column) if self.aggregate_fn else table_column
        return sql_col.label(self.label)

    def __repr__(self):
        return "SqlColumn(column_name=%s)" % self.column_name


class QueryMeta(object):
    def __init__(self, table_name, filters, group_by, distinct, order_by):
        self.filters = filters
        self.group_by = group_by
        self.distinct = distinct
        self.order_by = order_by
        self.table_name = table_name

    def append_column(self, view):
        pass

    def execute(self, connection, filter_values):
        raise NotImplementedError()

    def get_query_string(self, connection):
        raise NotImplementedError


class SimpleQueryMeta(QueryMeta):
    """
    Metadata about a query including the table being queried, list of columns, filters and group by columns.
    """
    def __init__(self, table_name, filters, group_by, distinct, order_by, start=None, limit=None):
        super(SimpleQueryMeta, self).__init__(table_name, filters, group_by, distinct, order_by)
        self.start = start
        self.limit = limit
        self.columns = []

    def append_column(self, column):
        self.columns.append(column.sql_column)

    def _check(self):
        if self.group_by:
            groups = list(self.group_by)
            for c in self.columns:
                if c.column_name in groups and not c.alias:
                    groups.remove(c.column_name)
                elif c.alias in groups:
                    groups.remove(c.alias)

            for g in groups:
                self.columns.append(SimpleSqlColumn(g, aggregate_fn=None, alias=g))

        labels = [col.label for col in self.columns]
        duplicates = [label for label, count in collections.Counter(labels).items() if count > 1]
        if duplicates:
            raise DuplicateColumnsException(
                'Query has duplicate columns: {}. '
                'Use aliases to disambiguate them.'.format(', '.join(duplicates))
            )

    def execute(self, connection, filter_values):
        query = self._build_query()
        return connection.execute(query, **filter_values).fetchall()

    def get_query_string(self, connection):
        query = self._build_query()
        return str(query.compile(connection))

    def count(self, connection, filter_values):
        assert self.start is None
        assert self.limit is None
        self._check()
        query = self._build_query_generic(self.columns, group_by=self.group_by, filters=self.filters,
                                          distinct=self.distinct)
        query = query.alias().count()
        return connection.execute(query, **filter_values).fetchall()[0][0]

    def totals(self, connection, filter_values, total_columns):
        assert self.start is None
        assert self.limit is None
        self._check()

        subquery = self._build_query_generic(self.columns, self.group_by, self.filters, self.distinct).alias()
        query = sqlalchemy.select().select_from(subquery)

        for total_column in total_columns:
            column = SimpleSqlColumn(total_column, sqlalchemy.func.sum)
            query.append_column(column.build_column())

        return dict(zip(
            total_columns,
            connection.execute(query, **filter_values).fetchall()[0]
        ))

    def _build_query(self):
        self._check()
        return self._build_query_generic(
            self.columns, self.group_by,
            self.filters, self.distinct, self.order_by, self.start, self.limit
        )

    def _build_query_generic(self, columns, group_by=None, filters=None, distinct=None,
                             order_by=None, start=None, limit=None):
        try:
            query = sqlalchemy.select()
            if group_by or distinct:
                cols = [c.column_name for c in columns]
                alias = [c.alias for c in columns]
                if distinct:
                    for col_key in distinct:
                        if col_key in cols:
                            query = query.distinct(column(col_key))
                        elif col_key in alias:
                            aliased_columns = [col.build_column() for col in columns if col.alias == col_key]
                            assert len(aliased_columns) == 1, "Only one column should have this alias"
                            query = query.distinct(aliased_columns[0])
                        else:
                            raise SqlAggException("Distinct on column not present in query columns or aliases")
                if group_by:
                    for group_key in group_by:
                        if group_key in cols:
                            query.append_group_by(column(group_key))
                        elif group_key in alias:
                            aliased_columns = [col.build_column() for col in columns if col.alias == group_key]
                            assert len(aliased_columns) == 1, "Only one column should have this alias"
                            query.append_group_by(aliased_columns[0])
                        else:
                            raise SqlAggException("Group by column not present in query columns or aliases")

            for c in columns:
                query.append_column(c.build_column())
        except KeyError as e:
            raise ColumnNotFoundException("Missing column in table (%s): %s" % (self.table_name, e))

        if filters:
            for filter in filters:
                query.append_whereclause(filter.build_expression())

        if not query.froms:
            query = query.select_from(table(self.table_name))

        if order_by:
            for order_by_column in order_by:
                order = order_by_column.build_expression()
                query = query.order_by(order)

        if start is not None:
            query = query.offset(start)
        if limit is not None:
            query = query.limit(limit)

        return query

    def __repr__(self):
        return "Querymeta(columns=%s, filters=%s, group_by=%s, distinct=%s, order_by=%s, table=%s)" % \
               (self.columns, self.filters, self.group_by, self.distinct, self.order_by, self.table_name)


class QueryContext(object):
    def __init__(self, table, filters=None, group_by=None, distinct=None, order_by=None, start=None, limit=None):
        self.table_name = table
        self.filters = filters or []
        self.group_by = group_by or []
        self.distinct = distinct or []
        self.order_by = order_by or []
        self.start = start
        self.limit = limit
        self.query_meta = {}

        if self.filters:
            assert all(isinstance(f, SqlFilter) for f in self.filters)

    def append_column(self, column):
        if isinstance(column, AliasColumn):
            return
        elif isinstance(column, AggregateColumn):
            for c in column.columns:
                self.append_column(c)
            return

        query_key = column.column_key
        query = self.query_meta.setdefault(query_key, self._new_query_meta(column))
        query.append_column(column)

    def _new_query_meta(self, column):
        if isinstance(column, QueryColumn):
            return column.get_query_meta(self.table_name, self.filters, self.group_by, self.order_by)
        else:
            table_name = column.table_name or self.table_name
            filters = column.filters or self.filters
            group_by = column.group_by or self.group_by
            order_by = column.order_by or self.order_by
            return SimpleQueryMeta(
                table_name, filters, group_by, self.distinct, order_by,
                start=self.start, limit=self.limit
            )

    def count(self, connection, filter_values=None):
        self.connection = connection
        query_meta_values = list(self.query_meta.values())
        if query_meta_values:
            return query_meta_values[0].count(
                connection, filter_values or {}
            )
        return 0

    def totals(self, connection, total_columns, filter_values=None):
        self.connection = connection
        query_meta_values = list(self.query_meta.values())
        if query_meta_values:
            return query_meta_values[0].totals(
                connection, filter_values or {}, total_columns
            )
        return {column: None for column in total_columns}

    def resolve(self, connection, filter_values=None):
        """
        Returns a dict containing the data of the following format:
        * If group_by == [] or None
            return a dict mapping column names to values: {'col_a': 1....}
        * If len(group_by) == 1
            return a dict mapping groupings to that group levels data
            e.g. {
                    'user1': {'user': 'user1', 'col_a': 1...},
                    'user2': {...}
                }
        * If len(group_by) > 1
            return a dict mapping tuple(groupsings) to that group levels data
            e.g. {
                    ('region1', 'subregion1'): {'region': 'region1', 'sub_region': 'subregion1', 'col_a': 1}
                    ('region1', 'subregion2'): {...}
                 }
        """
        self.connection = connection

        data = OrderedDict()
        for qm in self.query_meta.values():
            result = qm.execute(self.connection, filter_values or {})

            for r in result:
                if not qm.group_by:
                    row_key = None
                elif len(qm.group_by) == 1:
                    row_key = r[qm.group_by[0]]
                elif len(qm.group_by) > 1:
                    row_key = tuple([r[group] for group in qm.group_by])

                if qm.group_by:
                    if row_key is None:
                        # null values coming out of the database wreak havoc elsewhere in the code
                        row_key = ''
                    row = data.setdefault(row_key, {})
                    row.update(kvp for kvp in r.items())
                else:
                    data.update(kvp for kvp in r.items())

        return data

    def get_query_strings(self, connection):
        """Useful for debugging large queryies"""
        self.connection = connection
        return [
            qm.get_query_string(connection)
            for qm in self.query_meta.values()
        ]

    def __str__(self):
        return str(self.query_meta)


class SqlAggColumn(object):
    @property
    def name(self):
        raise NotImplementedError()

    @property
    def column_key(self):
        raise NotImplementedError()

    def get_value(self, row):
        raise NotImplementedError()


class QueryColumn(SqlAggColumn):
    def get_query_meta(self, table_name, filters, group_by, order_by):
        raise NotImplementedError()


class BaseColumn(SqlAggColumn):
    aggregate_fn = None

    def __init__(self, key, alias=None, table_name=None, filters=None, group_by=None, order_by=None):
        self.key = key
        self.alias = alias
        self.table_name = table_name
        self.filters = filters
        self.group_by = group_by
        self.order_by = order_by

        if self.filters:
            assert all(isinstance(f, SqlFilter) for f in self.filters)

        #TODO: allow 'having' e.g. count(x) having x > 4

    @property
    def name(self):
        return self.alias or self.key

    @property
    def column_key(self):
        return (
            self.table_name,
            tuple(sorted(self.filters)) if self.filters else None,
            tuple(self.group_by) if self.group_by else None
        )

    @property
    def sql_column(self):
        return SimpleSqlColumn(self.key, self.aggregate_fn, self.alias)

    def get_value(self, row):
        row_key = self.alias or self.key
        return row.get(row_key, None) if row else None


class CustomQueryColumn(BaseColumn, QueryColumn):
    query_cls = None
    name = None

    def get_query_meta(self, default_table_name, default_filters, default_group_by, default_order_by):
        table_name = self.table_name or default_table_name
        filters = self.filters or default_filters
        group_by = self.group_by or default_group_by
        order_by = self.order_by or default_order_by
        return self.query_cls(table_name, filters, group_by, order_by)

    @property
    def column_key(self):
        return (
            self.name, self.key, self.table_name,
            tuple(sorted(self.filters)) if self.filters else None,
            tuple(self.group_by) if self.group_by else None
        )


class AliasColumn(SqlAggColumn):
    """
    An AliasColumn doesn't contribute to the query. It is used to reference the value of an existing column.
    e.g. In an AggregateColumn
    """
    column_key = None

    def __init__(self, key):
        self.key = key

    @property
    def name(self):
        return self.key

    def get_value(self, row):
        return row.get(self.key, None) if row else None


class AggregateColumn(SqlAggColumn):
    def __init__(self, aggregate_fn, *columns):
        self.aggregate_fn = aggregate_fn
        self.columns = columns

    @property
    def name(self):
        return '_'.join([c.name for c in self.columns])

    def get_value(self, row):
        values = [v.get_value(row) for v in self.columns]
        return self.aggregate_fn(*values)



from __future__ import absolute_import
from __future__ import unicode_literals
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

tests_require=[
    'nose',
    'SQLAlchemy-Fixtures>=0.1.5',
    'fixture>=1.4',
    'psycopg2'
]


setup(
    name='sqlagg',
    version='0.16.0',
    description='SQL aggregation tool',
    author='Dimagi',
    author_email='dev@dimagi.com',
    url='http://github.com/dimagi/sql-agg',
    packages=['sqlagg'],
    license='MIT',
    install_requires=[
        'SQLAlchemy>=1.0.9',
    ],
    tests_require=tests_require,
    setup_requires=['nose'],
    extras_require={
        'test': tests_require,
    },
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ]
)
